import { Env } from "pi_utils/lang/env";
import { HttpRequest, HttpResponse } from 'pi_pt/init/httprpc'
import { cert_to_host_id, P2pMgrExport } from "pi_pt/native/pi_serv_lib/js_p2p";
import { utf8Encode } from "pi_utils/util/util";
import { File } from "pi_pt/native/pi_core_lib/js_file";
import { checkHttpPortAuth } from "./http_proxy.event";
import { asyncSpawnReceive } from "pi_pt/util/pid";
import { sendMsg } from "pi_pt/util/pid_rpc";
import { Tr } from "pi_utils/db/mgr";
import { consoleAll } from "pi_pt/util/util";
import { ClusterKey, readClusterAttr, writeClusterAttr, writeClusterInfo } from "pi_common/cluster/cluster_lib";
import { resultOk } from "../result_handle";
import { } from "pi_common/cluster/cluster_rpc_handle.event";

declare var env: Env;

// 证书转主机ID
// #[httpRpc=/mgr/cluster/cert_to_host_id]
export const certToHostId = (req: HttpRequest, _resp: HttpResponse) => {
    let _userData = checkHttpPortAuth(req);
    let cert = req.getData().cert;
    return resultOk(cert_to_host_id(utf8Encode(cert)));
}

// 获取默认配置
// #[httpRpc=/mgr/cluster/get_def_cfg]
export const getDefCfg = (req: HttpRequest, _resp: HttpResponse) => {
    let _userData = checkHttpPortAuth(req);
    return resultOk(P2pMgrExport.get_default_cfg());
}

// 证书上传
// #[httpRpc=/mgr/cluster/upload_cert]
export const uploadCert = async (req: HttpRequest, _resp: HttpResponse) => {
    let _userData = checkHttpPortAuth(req);
    const data = req.getData();
    const cert: string = data.cert;
    const ca = data.ca;
    const key = data.key;
    if (cert.length > 1000 || ca.length > 1000 || key.length > 1000) {
        throw "upload file length > 1000";
    }
    // 路径中不能包含上层路径
    if (data.path.indexOf("..") >= 0) {
        throw "path error ..";
    }
    const path = `cert/${data.path}`;
    let caPath = `${path}/ca_cert.pem`;
    let certPath = `${path}/cert.crt`;
    let keyPath = `${path}/key.pem`;

    await File.fs_mkdir(path);
    await File.fs_write_file(caPath, utf8Encode(ca));
    await File.fs_write_file(certPath, utf8Encode(cert));
    await File.fs_write_file(keyPath, utf8Encode(key));

    return resultOk({
        "ca": caPath,
        "cert": certPath,
        "key": keyPath
    });
}

// 打开终端
// #[httpRpc=/mgr/cluster/open_terminal]
export const openTerminal = async (req: HttpRequest, _resp: HttpResponse) => {
    // let _userData = checkHttpPortAuth(req);
    const data = req.getData();
    const cfg = data[""];
    // 创建处理pid
    const mgrPid = await asyncSpawnReceive("");
    await sendMsg(mgrPid, "cluster_open_terminal", cfg);
    await env.dbMgr.write(async (tr: Tr) => {
        await writeClusterInfo(tr, ClusterKey.TERMINAL_CFG, cfg);
        await writeClusterAttr(tr, ClusterKey.TERMINAL_PID, mgrPid);
    })
    return resultOk();
}

// 终端状态
// #[httpRpc=/mgr/cluster/terminal_state]
export const terminalState = async (req: HttpRequest, _resp: HttpResponse) => {
    // let _userData = checkHttpPortAuth(req);
    let mgrPid;
    await env.dbMgr.read(async (tr: Tr) => {
        mgrPid = await readClusterAttr(tr, ClusterKey.TERMINAL_PID);
    })
    if (!mgrPid) {
        throw "mgr pid not exist";
    }
    return resultOk(await sendMsg(mgrPid, "cluster_terminal_state", ""));
}

// 启动基础服务审核者
// #[httpRpc=/mgr/cluster/start_auditor]
export const startAuditor = async (req: HttpRequest, _resp: HttpResponse) => {
    // let _userData = checkHttpPortAuth(req);
    let mgrPid;
    await env.dbMgr.read(async (tr: Tr) => {
        mgrPid = await readClusterAttr(tr, ClusterKey.TERMINAL_PID);
    })
    if (!mgrPid) {
        throw "mgr pid not exist";
    }
    await sendMsg(mgrPid, "cluster_start_auditor", "");
    return resultOk();
}

// 启动基础服务申请者
// #[httpRpc=/mgr/cluster/start_applicant]
export const startApplicant = async (req: HttpRequest, _resp: HttpResponse) => {
    // let _userData = checkHttpPortAuth(req);
    let mgrPid;
    await env.dbMgr.read(async (tr: Tr) => {
        mgrPid = await readClusterAttr(tr, ClusterKey.TERMINAL_PID);
    })
    if (!mgrPid) {
        throw "mgr pid not exist";
    }
    await sendMsg(mgrPid, "cluster_start_applicant", "");
    return resultOk();
}


// 打开分布式配置服务
// #[httpRpc=/mgr/cluster/open_pi_config]
export const openPiConfig = async (req: HttpRequest, _resp: HttpResponse) => {
    // let _userData = checkHttpPortAuth(req);
    // TODO 判断集群状态
    // const data = req.getData();
    // const port = data.port;
    // const serviceName = data.name;
    // const mode = data.mode;
    let mgrPid;
    await env.dbMgr.read(async (tr: Tr) => {
        mgrPid = await readClusterAttr(tr, ClusterKey.TERMINAL_PID);
    })
    if (!mgrPid) {
        throw "mgr pid not exist";
    }
    await sendMsg(mgrPid, "cluster_open_cfg", "");
    return resultOk();
}

// 打开计算服务
// #[httpRpc=/mgr/cluster/open_calc]
export const openCalc = async (req: HttpRequest, _resp: HttpResponse) => {
    // let _userData = checkHttpPortAuth(req);
    // TODO 判断集群状态
    const data = req.getData();
    // const port = data.port;
    // const serviceName = data.name;
    // const mode = data.mode;
    let mgrPid;
    await env.dbMgr.read(async (tr: Tr) => {
        mgrPid = await readClusterAttr(tr, ClusterKey.TERMINAL_PID);
    })
    if (!mgrPid) {
        throw "mgr pid not exist";
    }
    await sendMsg(mgrPid, "cluster_open_calc", "");
    return resultOk();
}


// 打开代理服务
// #[httpRpc=/mgr/cluster/open_proxy]
export const openProxy = async (req: HttpRequest, _resp: HttpResponse) => {
    // let _userData = checkHttpPortAuth(req);
    // TODO 判断集群状态
    const data = req.getData();
    // const port = data.port;
    // const serviceName = data.name;
    // const mode = data.mode;
    let mgrPid;
    await env.dbMgr.read(async (tr: Tr) => {
        mgrPid = await readClusterAttr(tr, ClusterKey.TERMINAL_PID);
    })
    if (!mgrPid) {
        throw "mgr pid not exist";
    }
    await sendMsg(mgrPid, "cluster_net_proxy", "");
    return resultOk();
}