import { HttpRequest, HttpResponse } from "pi_pt/init/httprpc";
import { getSysEnv, get_pid, to_pid } from "pi_pt/util/util";
import { initProjectDef } from "pi_pt/init/load_project";
import { AuthDbNew, AuthDbKey, UserDb } from "management_page/server/db/user_db.struct";
import { Env } from "pi_utils/lang/env";
import * as User from "./user.event";
import { Session } from "pi_utils/net/session_new";
import { toByteArray } from "pi_utils/util/base64";
import { randomString } from "pi_utils/util/util";
import { sign, verify } from "pi_pt/util/sign";
import { resultErr, resultOk } from "../result_handle";
import { env_var } from "pi_pt/native/pi_serv_lib/js_env";

declare var env: Env;

// #[httpRpc=/mgr/api/porxy/login]
export const login = async (req: HttpRequest, resp: HttpResponse) => {
    let msg = req.getData();

    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    let localEnv = getEnv(msg.user);
    let r = await User.login(data, localEnv);
    console.log("login!!!!!!!!!!! r:", r);
    if (r.status === "success") {
        let token = randomString(6);
        // 设置会话信息
        sessionMgr.setSession(msg.user, { user: msg.user, session: localEnv.get("session"), token: { token, create_time: Date.now() } })
        r.data.token = token;
        r.data.user = msg.user;
        return r;
    }
    return r;
}

// 获取集群签名信息
// #[httpRpc=/mgr/api/porxy/cluster_sign]
export const clusterSign = async (req: HttpRequest, resp: HttpResponse) => {
    // 获取环境变量中的私钥
    let clusterKey = getSysEnv("CLUSTER_KEY");
    let userData = checkHttpPortAuth(req);
    let signObj = { timestamp: Date.now(), user: userData.user, auth: userData.session.get("auth") };
    let s = sign(signObj, clusterKey);

    return resultOk({ sign: s, msg: signObj });

}

// #[httpRpc=/mgr/api/porxy/cluster_login]
export const clusterLogin = async (req: HttpRequest, resp: HttpResponse) => {
    // 获取环境变量中的公钥
    let clusterPub = getSysEnv("CLUSTER_PUB");
    let msg = req.getData();
    let sginObj = JSON.parse(msg.msg);
    const now = Date.now();
    // 签名有效期10分钟
    if (sginObj.timestamp > now + (10 * 60 * 1000)) {
        return resultErr("sign timeout");
    }
    // 验证签名
    if (!verify(sginObj, msg.sign, clusterPub)) {
        return resultErr("sign error");
    }
    let localEnv = getEnv(sginObj.user);
    let token = randomString(6);
    let session: Session = localEnv.get("session");
    // 设置会话信息
    sessionMgr.setSession(sginObj.user, { user: sginObj.user, session: session, token: { token, create_time: Date.now() } })

    session.set("user", sginObj.user);
    session.set("auth", sginObj.auth);
    return {
        data: {
            token,
            user: sginObj.user
        }
    };
}

// #[httpRpc=/mgr/api/porxy/createUser]
export const createUser = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = getEnv(userData.user);
    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    return await User.createUser(data, localEnv);
}

// #[httpRpc=/mgr/api/porxy/delUser]
export const delUser = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = getEnv(userData.user);
    return await User.delUser(msg.user, localEnv);
}

// #[httpRpc=/mgr/api/porxy/mdfPwd]
export const mdfPwd = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let localEnv = getEnv(userData.user);
    let data = new UserDb();
    data.user = msg.user;
    data.pwd = msg.pwd;
    return await User.mdfPwd(data, localEnv);
}

// #[httpRpc=/mgr/api/porxy/users]
export const users = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    return await User.users(undefined, localEnv);
}

// #[httpRpc=/mgr/api/porxy/setAuth]
export const setAuth = async (req: HttpRequest, resp: HttpResponse) => {
    let userData = checkHttpPortAuth(req);
    let msg = req.getData();
    let json = JSON.parse(msg[""]);

    console.log("setAuth!!!!!!!!!!!!", json);
    let localEnv = getEnv(userData.user);
    let auth = new AuthDbNew();
    auth.user = json.user;
    auth.auth = json.auth;
    return await User.setAuth(auth, localEnv);
}




// 获取env
export const getEnv = (user: string): Env => {
    let localEnv = new Env();
    localEnv.dbMgr = env.dbMgr;
    localEnv.other = new Map();
    let session = sessionMgr.getSession(user) ? sessionMgr.getSession(user).session : new Session(Date.now().toString(), new Map(), () => { });

    localEnv.set("session", session);
    return localEnv;
}


const mapToJson = (map: Map<string, any>) => {
    let json = {};
    map.forEach((v, k) => {
        json[k] = v;
    })
    return json;
}

const jsonToMap = (json: object) => {
    let map = new Map();
    for (let k of Object.keys(json)) {
        console.log("jsonToMap:", { k, v: json[k] });
        map.set(k, json[k]);
    }
    return map;
}

const setCookie = (resp: HttpResponse, cookie: string) => {
    resp.setHeader("Set-Cookie", `attr=${cookie}`);
}

const getCookie = (req: HttpRequest): string => {
    let cookie = req.getHeader("cookie");
    return cookie
}

type UserData = {
    user: string,
    session: Session,
    token: {
        token: string,
        create_time: number
    },
}

class SessionMgr {
    sessions: Map<string, UserData>;
    constructor() {
        this.sessions = new Map();
    }

    setSession(user: string, userData: UserData) {
        this.sessions.set(user, userData);
    }

    getSession(user: string) {
        return this.sessions.get(user);
    }
}

let sessionMgr = new SessionMgr();


// httpPort权限检测
export const checkHttpPortAuth = (req: HttpRequest) => {
    let auth = req.getHeader("authorization");
    // Basic dGVzdDoxMjM0NTY=
    try {
        let txt = new TextDecoder().decode(toByteArray(auth.split(' ')[1]));
        const [username, password] = txt.split(':');
        let userData = sessionMgr.getSession(username);

        if (password !== userData.token.token || userData.token.create_time + 7 * 24 * 60 * 60 * 1000 < Date.now()) {
            throw new Error("check auth error");
        };
        return userData;
    } catch (error) {
        console.log("checkHttpPortAuth auth error:", error.message, error.stack);
        throw new Error("auth error");
    }
}