import { Env } from "pi_utils/lang/env";
import { Log } from "pi_pt/util/log";
import { BonBuffer } from "pi_utils/serialization/bon";
import { checkAuth } from "./user.event";
import { DBType, Item, Tr } from "pi_utils/db/mgr";
import { HttpRequest, HttpResponse } from 'pi_pt/init/httprpc'
import { TabDecode, TabEncode } from "pi_utils/serialization/tab_meta";
import { TabMeta } from "pi_utils/serialization/sinfo";
import { WebShellMgr } from "pi_pt/native/pi_serv_lib/js_webshell";
import { PT_ENV } from "pi_pt/constant";
import { MqttConnection } from "pi_pt/native/pi_serv_lib/js_net";
import { CWebShell, CWebShellOut } from "management_page/server/db/c_web_shell.struct";
import { publishEncodedMqttMessage } from "pi_pt/net/mqtt_broker";
import { OK_S } from "pi_utils/net/rpc_r.struct";
import { checkHttpBasicAuth } from "./auth_lib";
import { checkHttpPortAuth, getEnv } from "./http_proxy.event";
import { AUTH, FM, RESULT_NUM } from "../../../constant";

declare var ESProcess;
declare var env: Env;

// -------------------客户端mqtt调用------------------------
// 注册web_shell
// #[rpc=rpcServer, src=*] 
export const regWebShell = async (rid) => {
    let mc: MqttConnection = env.get(PT_ENV.mqttSession);
    const clientId = mc.client_id();
    let cws = new CWebShell();
    cws.client_id = clientId;
    cws.session_pid = ESProcess.pid.toString();
    cws.rid = Number(rid);
    await env.dbMgr.write(async (tr: Tr) => {
        await tr.modify([{ ware: "memory", tab: CWebShell._$info.name, key: cws.rid, value: cws }], 1000, true);
    })

    // 返回topic
    let topic = `topic/mgr/webshell/${clientId}`;
    return topic;
};

// 输出上报
// #[rpc=rpcServer, src=*] 
export const writeOut = async (out: string) => {
    let mc: MqttConnection = env.get(PT_ENV.mqttSession);
    const clientId = mc.client_id();
    console.log('writeOut=====', clientId, out);
    await env.dbMgr.write(async (tr: Tr) => {
        let cWebOut: CWebShellOut = (await tr.query([{ ware: "memory", tab: CWebShellOut._$info.name, key: clientId }], 1000, false))[0].value;

        if (!cWebOut) {
            cWebOut = new CWebShellOut();
            cWebOut.client_id = clientId;
            cWebOut.outs = [];
        }
        cWebOut.outs.push(out);
        await tr.modify([{ ware: "memory", tab: CWebShellOut._$info.name, key: cWebOut.client_id, value: cWebOut }], 1000, true);
    })

    return "ok"
}

// ------------------管理端web调用-----------------------------------------------------
/**
 * 获取web_shell列表
 */
// #[httpRpc=/mgr/cwebshell/list]
export const list = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.CLIENT_WEB_SHELL, AUTH.R, localEnv)) throw "权限错误";
    let lists = [];
    await env.dbMgr.read(async (tr: Tr) => {
        for await (const webShell of tr.kvList("memory", CWebShell._$info.name, null, false, "")) {
            lists.push(webShell);
        }
    })
    return JSON.stringify(lists);
}

/**
 * 关闭web_shell
 */
// #[httpRpc=/mgr/cwebshell/close]
export const close = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.CLIENT_WEB_SHELL, AUTH.R, localEnv)) throw "权限错误";

    let data = req.getData();
    let rid = data.rid;
    await env.dbMgr.write(async (tr: Tr) => {
        await tr.modify([{ ware: "memory", tab: CWebShell._$info.name, key: rid }], 1000, true);
    })
    return "ok";
}

/**
 * 发送指令
 */
// #[httpRpc=/mgr/cwebshell/send_cmd]
export const sendCmd = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.CLIENT_WEB_SHELL, AUTH.W, localEnv)) throw "权限错误";

    let data = JSON.parse(req.getData()[""]);
    let clientId = data.client_id;
    let cmd = data.cmd;
    let ok = new OK_S();
    console.log("!!!!!!cmd:", cmd);
    ok.value = typeof cmd === 'string' ? cmd : JSON.stringify(cmd);
    // 推送指令到topic
    let msg2 = new BonBuffer().write(ok).getBuffer();
    // 发送到客户端
    publishEncodedMqttMessage('*', true, `topic/mgr/webshell/${clientId}`, msg2);

    return "ok";
}

/**
 * 读取webshell输出
 */
// #[httpRpc=/mgr/cwebshell/read_out]
export const readOut = async (req: HttpRequest, resp: HttpResponse) => {
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.CLIENT_WEB_SHELL, AUTH.R, localEnv)) throw "权限错误";

    let data = req.getData();
    let clientId = data.client_id;
    let outs;
    await env.dbMgr.write(async (tr: Tr) => {
        let cWebOut: CWebShellOut = (await tr.query([{ ware: "memory", tab: CWebShellOut._$info.name, key: clientId }], 1000, false))[0].value;
        if (!cWebOut) {
            cWebOut = new CWebShellOut();
            cWebOut.client_id = clientId;
            cWebOut.outs = [];
        }
        outs = cWebOut.outs;
        // 清空缓存
        cWebOut.outs = [];
        await tr.modify([{ ware: "memory", tab: CWebShellOut._$info.name, key: cWebOut.client_id, value: cWebOut }], 1000, true);
    });
    return outs;
}
/**
 * 启动webshell 保证 pi_common/client_webshell/webshell有startWebShell这个方法
 * http://127.0.0.1:8099/mgr/cwebshell/send_command?rid=100005
 */
// #[httpRpc=/mgr/cwebshell/start_shell]
export const startShell = async (req: HttpRequest, resp: HttpResponse) => {
    console.log(req.getData());
    // 检测授权信息
    // checkHttpBasicAuth(req, resp);
    let userData = checkHttpPortAuth(req);
    let localEnv = getEnv(userData.user);
    if (!checkAuth(FM.CLIENT_WEB_SHELL, AUTH.W, localEnv)) throw "权限错误";
    return await env.dbMgr.read(async (tr: Tr) => {
        let shell: CWebShell = (await tr.query([{ ware: "memory", tab: CWebShell._$info.name, key: Number(req.getData().rid) }], 1000, false))[0].value;
        if (!shell) return "";
        // pi_common_tools\client_webshell\webshell
        // ESProcess.send(Number(shell.session_pid), ["cmd", "_$pi._modules.get('pi_common/client_webshell/webshell').exports.startWebShell()"]);
        ESProcess.send(Number(shell.session_pid), ["cmd", "_$pi._modules.get('pi_common_tools/client_webshell/webshell').exports.startWebShell()"]);
        return shell.client_id;
    })
}




